package Javascript::PurePerl;
use strict;

use vars qw($VERSION);

$VERSION = '0.01';

use Parse::RecDescent;
use base 'Class::Accessor';

__PACKAGE__->mk_accessors(qw( tree ));

my $grammar = q<
  <autotree>

  Program: <skip: qr{\s* (/[*] .*? [*]/ \s*)*}x> SourceElement(s) /^\Z/

  SourceElement: FunctionDeclaration
               | Statement

  FormalParameterList: Identifier(s /,/)

  FunctionDeclaration: 'function' Identifier '(' FormalParameterList(?) ')' '{' FunctionBody '}'

  FunctionExpression: 'function' Identifier(?) '(' FormalParameterList(?) ')' '{' FunctionBody '}'

  FunctionBody: SourceElement(s /;/)

  Identifier: /[a-zA-Z]\w*/

  Expression: AssignmentExpression(s /,/)

  AssignmentExpression: ConditionalExpression
                      | LeftHandSideExpression AssignmentOperator AssignmentExpression
                      # /\d+/
  AssignmentOperator: /^=(?!=)/
                    | '*='
                    | '/='
                    | '%='
                    | '+='
                    | '-='
                    | '\<\<='
                    | '\>\>='
                    | '\>\>\>='
                    | '&='
                    | '^='
                    | '|='

  ConditionalExpression: LogicalORExpression ('?' AssignmentExpression ':' AssignmentExpression)(?)

  LogicalORExpression: (LogicalANDExpression '||')(s?) LogicalANDExpression

  LogicalANDExpression: (BitwiseORExpression '&&')(s?) BitwiseORExpression

  BitwiseORExpression: (BitwiseXORExpression BitwiseOROperator)(s?) BitwiseXORExpression
  BitwiseOROperator: /\\|(?!\\|)/

  BitwiseXORExpression: (BitwiseANDExpression '^')(s?) BitwiseANDExpression

  BitwiseANDExpression: (EqualityExpression '&')(s?) EqualityExpression

  EqualityExpression: (RelationalExpression EqualityOperator)(s?) RelationalExpression
  EqualityOperator: '=='
                  | '!='
                  | '==='
                  | '!=='

  RelationalExpression: (ShiftExpression RelationalOperator)(s?) ShiftExpression
  RelationalOperator: '\<='
                    | '\>='
                    | '\<'
                    | '\>'
                    | 'instanceof'
                    | 'in'

  ShiftExpression: (AdditiveExpression ShiftOperator)(s?) AdditiveExpression
  ShiftOperator: /\<\<(?!=)/
               | /\>\>\>(?!=)/
               | /\>\>(?!=)/

  AdditiveExpression: (MultiplicativeExpression AdditiveOperator)(s?) MultiplicativeExpression
  AdditiveOperator: /[+](?![+])/
                  | /[-](?![-])/

  MultiplicativeExpression: (UnaryExpression MultiplicativeOperator)(s?) UnaryExpression
  MultiplicativeOperator: '*'
                        | '/'
                        | '%'

  UnaryExpression: PostfixExpression
                 | UnaryPrefixOperator UnaryExpression
  UnaryPrefixOperator: 'delete'
                     | 'void'
                     | 'typeof'
                     | '++'
                 	   | '--'
                     | /[+](?![+])/
                     | /[-](?![-])/
                     | '~'
                     | '!'

  PostfixExpression: LeftHandSideExpression PostfixOperator(?)
  PostfixOperator: '++'
                 | '--'

  LeftHandSideExpression: NewExpression
                        | CallExpression

  # The standard says:
  #   NewExpression: 'new' NewExpression
  # but that makes not much sense currently
  NewExpression: 'new' QualifiedMemberExpression Arguments(?)
               | QualifiedMemberExpression

  Arguments: '(' ')'

  QualifiedMemberExpression: (MemberExpression)(s) (IndexedAccess|MemberAccess)(?)
  IndexedAccess: '[' Expression ']'
  MemberAccess: '.' Identifier

  MemberExpression: PrimaryExpression
                  | FunctionExpression
                  # | 'new' MemberExpression Arguments
                  # see above

  CallExpression: IntegerExpression

  PrimaryExpression: IntegerExpression

  IntegerExpression: /^\d+/

  Initializer: '=' AssignmentExpression

  Statement: Block
           | VariableStatement
           | EmptyStatement
           | ExpressionStatement
           #| IfStatement
           #| IterationStatement
           #| ContinueStatement
           #| BreakStatement
           #| ReturnStatement
           #| WithStatement
           #| LabelledStatement
           #| SwitchStatement
           #| ThrowStatement
           #| TryStatement

  Block: '{' Statement(s?) '}'
  VariableStatement: 'var' <commit> VariableDeclarationList ';'
                   | <error>
  EmptyStatement: ';'
  ExpressionStatement: ...!'function' Expression ';'
                     | ...!'{'        Expression ';'
  #IfStatement: 'if' '(' Expression ')' 'then' Statement 'else' Statement
  #           | 'if' '(' Expression ')' 'then' Statement
  #IterationStatement: 'for' /[^;]*/ ';'
  #ContinueStatement: 'continue' /[^;]*/ ';'
  #BreakStatement: 'break' /[^;]*/ ';'
  #ReturnStatement: 'return' /[^;]*/ ';'
  #WithStatement: 'with' /[^;]*/ ';'
  #LabelledStatement: /[^;]*/ ';'
  #SwitchStatement: 'switch' /[^;]*/ ';'
  #ThrowStatement: 'throw' /[^;]*/ ';'
  #TryStatement: 'try' /[^;]*/ 'catch' 'finally' ';'

  VariableDeclarationList: VariableDeclaration(s /,/)

  VariableDeclaration: Identifier Initializer(?)

>;

=for later
=cut

sub new {
  my ($class,%args) = @_;
  my $self = {};
  bless $self, $class;

  if ($args{code}) {
    $self->parse(code => $args{code});
  };

  $self;
};

sub parse {
  my ($self,%args) = @_;
  if (! $self->{tree}) {
    my $parser = Parse::RecDescent->new($grammar);
    $self->tree( $parser->Program($args{code}));
    $self->optimize();
  };
  make_document( $self->tree );
};

sub optimize {
  my $self = shift;
};

sub append_subtree {
  my ($document,$parent,$tree) = @_;
  croak "Need a document" unless $document;
  croak "Need a parent element" unless $parent;

  if (ref $tree eq 'ARRAY') {
    append_subtree($document,$parent,$_)
      for @$tree;
  elsif ( $tree->{__VALUE__} ) {

  } else {
    my ($node,$element_name);
    if ($tree->{__RULE__}) {
      $element_name = $tree->{__RULE__};
      $node = $document->createElement( $element_name );
    } else {
      $element_name = ref $tree;
      $node = $document->createElement( $element_name );
      $node->appendChild( $document->createTextNode( $tree->{__VALUE__} ));
    };

    my @children = map { $tree->{$_} } grep {   ! /^__/
                                             && ! exists $tree->{"$_(s)"}} keys %$tree;

    $parent->appendChild( $node );
    append_subtree($document,$node,$_)
      for @children;
  };
};

sub make_document {
  my $tree = shift;

  my $document = XML::LibXML::Document->createDocument(1,'');
  $document->setStandalone(1);
  my $root = $document->createElement('AST');
  $document->setDocumentElement($root);
  append_subtree($document,$root,$tree);

  return $document;
};

1;
__END__

=head1 NAME

Javascript::PurePerl - Javascript Parser

=head1 SYNOPSIS

=for example begin

  use Javascript::PurePerl;

  my $tree = Javascript::PurePerl->new->parse($js);

=for example end

=head1 DESCRIPTION

Simplicistic top-down parser for Javascript

=head2 EXPORT

None by default.

=head1 SEE ALSO

L<Parse::RecDescent>, L<Javascript>

=head1 AUTHOR

Max Maischein E<lt>corion@cpan.orgE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2003,2004 by Max Maischein

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself, either Perl version 5.8.2 or,
at your option, any later version of Perl 5 you may have available.

=cut
